/***
*
*  Copyright (c) 1999, Valve LLC. All rights reserved.
*
*  This product contains software technology licensed from Id
*  Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc.
*  All Rights Reserved.
*
*   Use, distribution, and modification of this source code and/or resulting
*   object code is restricted to non-commercial enhancements to products from
*   Valve LLC.  All other use, distribution, or modification is prohibited
*   without written permission from Valve LLC.
*
****/

//
// HPB_bot - botman's High Ping Bastard bot
//
// (http://planethalflife.com/botman/)
//
// util.cpp
//

#include "extdll.h"
#include "util.h"
#include "engine.h"
#include "cbase.h"


int gmsgTextMsg = 0;
int gmsgSayText = 0;
int gmsgShowMenu = 0;


Vector UTIL_VecToAngles( const Vector &vec )
{
   float rgflVecOut[3];
   VEC_TO_ANGLES(vec, rgflVecOut);
   return Vector(rgflVecOut);
}


// Overloaded to add IGNORE_GLASS
void UTIL_TraceLine( const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, IGNORE_GLASS ignoreGlass, edict_t *pentIgnore, TraceResult *ptr )
{
   TRACE_LINE( vecStart, vecEnd, (igmon == ignore_monsters ? TRUE : FALSE) | (ignoreGlass?0x100:0), pentIgnore, ptr );
}


void UTIL_TraceLine( const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, edict_t *pentIgnore, TraceResult *ptr )
{
   TRACE_LINE( vecStart, vecEnd, (igmon == ignore_monsters ? TRUE : FALSE), pentIgnore, ptr );
}


void UTIL_MakeVectors( const Vector &vecAngles )
{
   MAKE_VECTORS( vecAngles );
}


/*
edict_t *UTIL_FindEntityInSphere( edict_t *pentStart, const Vector &vecCenter, float flRadius )
{
   edict_t  *pentEntity;

   pentEntity = FIND_ENTITY_IN_SPHERE( pentStart, vecCenter, flRadius);

   if (!FNullEnt(pentEntity))
      return pentEntity;

   return NULL;
}


edict_t *UTIL_FindEntityByString( edict_t *pentStart, const char *szKeyword, const char *szValue )
{
   edict_t *pentEntity;

   pentEntity = FIND_ENTITY_BY_STRING( pentStart, szKeyword, szValue );

   if (!FNullEnt(pentEntity))
      return pentEntity;
   return NULL;
}

edict_t *UTIL_FindEntityByClassname( edict_t *pentStart, const char *szName )
{
   return UTIL_FindEntityByString( pentStart, "classname", szName );
}

edict_t *UTIL_FindEntityByTargetname( edict_t *pentStart, const char *szName )
{
   return UTIL_FindEntityByString( pentStart, "targetname", szName );
}
*/


int UTIL_PointContents( const Vector &vec )
{
   return POINT_CONTENTS(vec);
}


void UTIL_SetSize( entvars_t *pev, const Vector &vecMin, const Vector &vecMax )
{
   SET_SIZE( ENT(pev), vecMin, vecMax );
}


void UTIL_SetOrigin( entvars_t *pev, const Vector &vecOrigin )
{
   SET_ORIGIN(ENT(pev), vecOrigin );
}


void ClientPrint( entvars_t *client, int msg_dest, const char *msg_name, const char *param1, const char *param2, const char *param3, const char *param4 )
{
   if (gmsgTextMsg == 0)
      gmsgTextMsg = REG_USER_MSG( "TextMsg", -1 );

	MESSAGE_BEGIN( MSG_ONE, gmsgTextMsg, NULL, client );
		WRITE_BYTE( msg_dest );
		WRITE_STRING( msg_name );

		if ( param1 )
			WRITE_STRING( param1 );
		if ( param2 )
			WRITE_STRING( param2 );
		if ( param3 )
			WRITE_STRING( param3 );
		if ( param4 )
			WRITE_STRING( param4 );

	MESSAGE_END();
}

void UTIL_SayText( const char *pText, edict_t *pEdict )
{
   if (gmsgSayText == 0)
      gmsgSayText = REG_USER_MSG( "SayText", -1 );

   pfnMessageBegin( MSG_ONE, gmsgSayText, NULL, pEdict );
      pfnWriteByte( ENTINDEX(pEdict) );
      pfnWriteString( pText );
   pfnMessageEnd();
}

void UTIL_HudMessage( CBaseEntity *pEntity, const hudtextparms_t &textparms, const char *pMessage )
{
	if ( !pEntity || !pEntity->IsNetClient() )
		return;

	MESSAGE_BEGIN( MSG_ONE, SVC_TEMPENTITY, NULL, pEntity->edict() );
		WRITE_BYTE( TE_TEXTMESSAGE );
		WRITE_BYTE( textparms.channel & 0xFF );

		WRITE_SHORT( FixedSigned16( textparms.x, 1<<13 ) );
		WRITE_SHORT( FixedSigned16( textparms.y, 1<<13 ) );
		WRITE_BYTE( textparms.effect );

		WRITE_BYTE( textparms.r1 );
		WRITE_BYTE( textparms.g1 );
		WRITE_BYTE( textparms.b1 );
		WRITE_BYTE( textparms.a1 );

		WRITE_BYTE( textparms.r2 );
		WRITE_BYTE( textparms.g2 );
		WRITE_BYTE( textparms.b2 );
		WRITE_BYTE( textparms.a2 );

		WRITE_SHORT( FixedUnsigned16( textparms.fadeinTime, 1<<8 ) );
		WRITE_SHORT( FixedUnsigned16( textparms.fadeoutTime, 1<<8 ) );
		WRITE_SHORT( FixedUnsigned16( textparms.holdTime, 1<<8 ) );

		if ( textparms.effect == 2 )
			WRITE_SHORT( FixedUnsigned16( textparms.fxTime, 1<<8 ) );
		
		if ( strlen( pMessage ) < 512 )
		{
			WRITE_STRING( pMessage );
		}
		else
		{
			char tmp[512];
			strncpy( tmp, pMessage, 511 );
			tmp[511] = 0;
			WRITE_STRING( tmp );
		}
	MESSAGE_END();
}

void UTIL_HudMessageAll( const hudtextparms_t &textparms, const char *pMessage )
{
	int			i;

	for ( i = 1; i <= gpGlobals->maxClients; i++ )
	{
		CBaseEntity *pPlayer = UTIL_PlayerByIndex( i );
		if ( pPlayer )
			UTIL_HudMessage( pPlayer, textparms, pMessage );
	}
}




#ifdef   DEBUG
edict_t *DBG_EntOfVars( const entvars_t *pev )
{
   if (pev->pContainingEntity != NULL)
      return pev->pContainingEntity;
   ALERT(at_console, "entvars_t pContainingEntity is NULL, calling into engine");
   edict_t* pent = (*g_engfuncs.pfnFindEntityByVars)((entvars_t*)pev);
   if (pent == NULL)
      ALERT(at_console, "DAMN!  Even the engine couldn't FindEntityByVars!");
   ((entvars_t *)pev)->pContainingEntity = pent;
   return pent;
}
#endif //DEBUG





bool IsAlive(edict_t *pEdict)
{
   return ((pEdict->v.deadflag == DEAD_NO) &&
           (pEdict->v.health > 0) && (pEdict->v.movetype != MOVETYPE_NOCLIP));
}


bool FInViewCone(Vector *pOrigin, edict_t *pEdict)
{
   Vector2D vec2LOS;
   float    flDot;

   UTIL_MakeVectors ( pEdict->v.angles );

   vec2LOS = ( *pOrigin - pEdict->v.origin ).Make2D();
   vec2LOS = vec2LOS.Normalize();

   flDot = DotProduct (vec2LOS , gpGlobals->v_forward.Make2D() );

   if ( flDot > 0.50 )  // 60 degree field of view 
   {
      return TRUE;
   }
   else
   {
      return FALSE;
   }
}


bool FVisible( const Vector &vecOrigin, edict_t *pEdict )
{
   TraceResult tr;
   Vector      vecLookerOrigin;

   // look through caller's eyes
   vecLookerOrigin = pEdict->v.origin + pEdict->v.view_ofs;

   int bInWater = (UTIL_PointContents (vecOrigin) == CONTENTS_WATER);
   int bLookerInWater = (UTIL_PointContents (vecLookerOrigin) == CONTENTS_WATER);

   // don't look through water
   if (bInWater != bLookerInWater)
      return FALSE;

   UTIL_TraceLine(vecLookerOrigin, vecOrigin, ignore_monsters, ignore_glass, pEdict, &tr);

   if (tr.flFraction != 1.0)
   {
      return FALSE;  // Line of sight is not established
   }
   else
   {
      return TRUE;  // line of sight is valid.
   }
}


Vector GetGunPosition(edict_t *pEdict)
{
   return (pEdict->v.origin + pEdict->v.view_ofs);
}



Vector VecBModelOrigin(edict_t *pEdict)
{
   return pEdict->v.absmin + (pEdict->v.size * 0.5);
}


void UTIL_ShowMenu( edict_t *pEdict, int slots, int displaytime, bool needmore, char *pText )
{
   if (gmsgShowMenu == 0)
      gmsgShowMenu = REG_USER_MSG( "ShowMenu", -1 );

   pfnMessageBegin( MSG_ONE, gmsgShowMenu, NULL, pEdict );

   pfnWriteShort( slots );
   pfnWriteChar( displaytime );
   pfnWriteByte( needmore );
   pfnWriteString( pText );

   pfnMessageEnd();
}


// *************************
// ****** NeoTF Stuff ******
// *************************

// void UTIL_TraceLine(const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, IGNORE_GLASS ignoreGlass, edict_t *pentIgnore, TraceResult *ptr)
// int Reflect is the reflection line that is actually to be traced, thus a number of calls are required
void UTIL_TraceReflect(const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, IGNORE_GLASS ignoreGlass, edict_t *pentIgnore, TraceResult *ptr, int Reflect)
{
	Vector tVector, tStart, tEnd;
    TraceResult ttr;
    int counter;

    tStart = vecStart;
    tEnd = vecEnd;
    ttr = *ptr;
    
    for(counter = 0;counter<Reflect;counter++)
    {
    	UTIL_TraceLine(tStart, tEnd, igmon, ignoreGlass, pentIgnore, &ttr);
		tEnd = ((tEnd-tStart).Length())*((2*DotProduct(-(tEnd-tStart).Normalize(),ttr.vecPlaneNormal)*ttr.vecPlaneNormal)+(tEnd-tStart).Normalize());
        tStart = ttr.vecEndPos;
   		if (ttr.flFraction >= 1.0)
        	break;
    }

    *ptr = ttr;
}

void UTIL_TraceReflect(const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, edict_t *pentIgnore, TraceResult *ptr, int Reflect)
{
	Vector tVector, tStart, tEnd;
    TraceResult ttr;
    int counter;

    tStart = vecStart;
    tEnd = vecEnd;
    ttr = *ptr;

    for(counter = 0;counter<Reflect;counter++)
    {
    	UTIL_TraceLine(tStart, tEnd, igmon, pentIgnore, &ttr);
		tEnd = ((tEnd-tStart).Length())*((2*DotProduct(-(tEnd-tStart).Normalize(),ttr.vecPlaneNormal)*ttr.vecPlaneNormal)+(tEnd-tStart).Normalize());
        tStart = ttr.vecEndPos;
   		if (ttr.flFraction >= 1.0)
        	break;
    }

    *ptr = ttr;
}
//util team color - by RedFox of FoxBot - http://www.foxbot.net 
int UTIL_GetTeamColor(edict_t *pEntity) 
{ 
	//crash chack here 
	if(pEntity==NULL) 
	return -1; 
		{ 
		char *infobuffer; 
		char topcolor[32]; 

		infobuffer = (*g_engfuncs.pfnGetInfoKeyBuffer)( pEntity ); 
		strcpy(topcolor, (g_engfuncs.pfnInfoKeyValue(infobuffer, "topcolor"))); 
		//infobuffer = pfnGetInfoKeyBuffer( pEntity ); 
		//strcpy(topcolor, pfnInfoKeyValue(infobuffer, \"topcolor\")); 

		//used for spy checking 
		if(strcmp(topcolor,"150")==0 || strcmp(topcolor,"153")==0 || strcmp(topcolor,"148")==0 || strcmp(topcolor,"140")==0) return 1;//blue 
		if(strcmp(topcolor,"250")==0 || strcmp(topcolor,"255")==0 || strcmp(topcolor,"5")==0) return 2;//red 
		if(strcmp(topcolor,"45")==0) return 3;//yellow 
		if(strcmp(topcolor,"100")==0 || strcmp(topcolor,"80")==0) return 4;//green 

		//may cause crash if human player?? 
		return pEntity->v.team; // TFC teams are 1-4 based 
		} 
	return -1; 
}

int UTIL_IsMasterTriggered(string_t sMaster, CBaseEntity *pActivator)
{
	if (sMaster)
	{
		edict_t *pentTarget = FIND_ENTITY_BY_TARGETNAME(NULL, STRING(sMaster));
	
		if ( !FNullEnt(pentTarget) )
		{
			CBaseEntity *pMaster = CBaseEntity::Instance(pentTarget);
			if ( pMaster && (pMaster->ObjectCaps() & FCAP_MASTER) )
				return pMaster->IsTriggered( pActivator );
		}

		ALERT(at_console, "Master was null or not a master!\n");
	}

	// if this isn't a master entity, just say yes.
	return 1;
}


char *UTIL_StringSearchReplace(const char *Search, const char *Replace, const char *Input)
{
	static char Output[8192], Temp[8192];
    char *ptr;
    sprintf(Output, "%s", Input);
    while(ptr = strstr(Output, Search))
    {
        *ptr = '\0';
    	sprintf(Temp, "%s%s%s", Output, Replace, ptr + strlen(Search));
        sprintf(Output, "%s", Temp);
    }
    return Output;
}

bool UTIL_CheckBox(Vector Origin, Vector Dimensions)
{
	// This function checks up, down, left, right, front and back in Dimensions/2 units in each direction from the origin. It's for establishing
    // whether it's OK for items to be placed in a certain position, but it isn't foolproof -- doesn't check diagolnals or slope of land etc.

    TraceResult tr;
    UTIL_TraceLine(Origin, Origin + Vector(Dimensions.x/2, 0, 0), dont_ignore_monsters, NULL, &tr);
    if(tr.flFraction!=1.0) return 0;
    UTIL_TraceLine(Origin, Origin - Vector(Dimensions.x/2, 0, 0), dont_ignore_monsters, NULL, &tr);
    if(tr.flFraction!=1.0) return 0;
    UTIL_TraceLine(Origin, Origin + Vector(0, Dimensions.y/2, 0), dont_ignore_monsters, NULL, &tr);
    if(tr.flFraction!=1.0) return 0;
    UTIL_TraceLine(Origin, Origin - Vector(0, Dimensions.y/2, 0), dont_ignore_monsters, NULL, &tr);
    if(tr.flFraction!=1.0) return 0;
    UTIL_TraceLine(Origin, Origin + Vector(0, 0, Dimensions.z/2), dont_ignore_monsters, NULL, &tr);
    if(tr.flFraction!=1.0) return 0;
    UTIL_TraceLine(Origin, Origin - Vector(0, 0, Dimensions.z/2), dont_ignore_monsters, NULL, &tr);
    if(tr.flFraction!=1.0) return 0;

    return 1;
}

float UTIL_TossAngleFromSpeed(Vector Origin, Vector Target, float Speed)
{
	// Projectile motion equation:
    // y = utsin(theta) - (gt^2)/2
    // x = utcos(theta)
    // y = xtan(theta) - (gx^2)/(2(v^2)(cos(theta)^2))

    // NOTE: I wasn't able to solve for angle knowing velocity and position. Could be unsolvable.

    float x = sqrt((Target.x - Origin.x)*(Target.x - Origin.x) + (Target.y - Origin.y)*(Target.y - Origin.y));
    float y = Target.z - Origin.z;
    float v = Speed;
}

float UTIL_TossSpeedFromAngle(Vector Origin, Vector Target, float Theta)
{
	float speed;
    float speedsq;
	float Alpha = atan((Target.z - Origin.z) / sqrt((Target.x - Origin.x)*(Target.x - Origin.x) + (Target.y - Origin.y)*(Target.y - Origin.y)));
    // ALERT(at_console, "> From UTIL: Theta %d\n", (int)(Theta*180/M_PI));
    if((Theta + Alpha) >= M_PI/2) Theta = 0;
    speedsq = ((Target - Origin).Length() * CVAR_GET_FLOAT("sv_gravity") * 0.65 * pow(cos(Alpha),2)) / (2 * sin(Theta) * cos(Theta + Alpha));
    // ALERT(at_console, "> From UTIL: speedsq %d\n", (int)speedsq);
    if(speedsq >= 0)
    {
    	speed = sqrt(speedsq);
        return speed;
    }
    else
    {
    	// ALERT(at_console, "> ERROR: Negative speedsq in Toss\n");
		speed = sqrt(-speedsq);
	    return speed;
    }
}

// ********************************
// ****** End of NeoTF Stuff ******
// ********************************


CBaseEntity *UTIL_FindEntityInSphere( CBaseEntity *pStartEntity, const Vector &vecCenter, float flRadius )
{
	edict_t	*pentEntity;

	if (pStartEntity)
		pentEntity = pStartEntity->edict();
	else
		pentEntity = NULL;

	pentEntity = FIND_ENTITY_IN_SPHERE( pentEntity, vecCenter, flRadius);

	if (!FNullEnt(pentEntity))
		return CBaseEntity::Instance(pentEntity);
	return NULL;
}


CBaseEntity *UTIL_FindEntityByString( CBaseEntity *pStartEntity, const char *szKeyword, const char *szValue )
{
	edict_t	*pentEntity;

	if (pStartEntity)
		pentEntity = pStartEntity->edict();
	else
		pentEntity = NULL;

	pentEntity = FIND_ENTITY_BY_STRING( pentEntity, szKeyword, szValue );

	if (!FNullEnt(pentEntity))
		return CBaseEntity::Instance(pentEntity);
	return NULL;
}

CBaseEntity *UTIL_FindEntityByClassname( CBaseEntity *pStartEntity, const char *szName )
{
	return UTIL_FindEntityByString( pStartEntity, "classname", szName );
}

CBaseEntity *UTIL_FindEntityByTargetname( CBaseEntity *pStartEntity, const char *szName )
{
	return UTIL_FindEntityByString( pStartEntity, "targetname", szName );
}
////////////////////////AUS/////////////////////////////////////
//CBaseEntity *UTIL_FindEntityByNetname( CBaseEntity *pStartEntity, const char *szName )
//{
//	return UTIL_FindEntityByString( pStartEntity, "netname", szName );
//}
////////////////////////AUS/////////////////////////////////////


CBaseEntity *UTIL_FindEntityGeneric( const char *szWhatever, Vector &vecSrc, float flRadius )
{
	CBaseEntity *pEntity = NULL;

	pEntity = UTIL_FindEntityByTargetname( NULL, szWhatever );
	if (pEntity)
		return pEntity;

	CBaseEntity *pSearch = NULL;
	float flMaxDist2 = flRadius * flRadius;
	while ((pSearch = UTIL_FindEntityByClassname( pSearch, szWhatever )) != NULL)
	{
		float flDist2 = (pSearch->pev->origin - vecSrc).Length();
		flDist2 = flDist2 * flDist2;
		if (flMaxDist2 > flDist2)
		{
			pEntity = pSearch;
			flMaxDist2 = flDist2;
		}
	}
	return pEntity;
}


// returns a CBaseEntity pointer to a player by index.  Only returns if the player is spawned and connected
// otherwise returns NULL
// Index is 1 based
CBaseEntity	*UTIL_PlayerByIndex( int playerIndex )
{
	CBaseEntity *pPlayer = NULL;

	if ( playerIndex > 0 && playerIndex <= gpGlobals->maxClients )
	{
		edict_t *pPlayerEdict = INDEXENT( playerIndex );
		if ( pPlayerEdict && !pPlayerEdict->free )
		{
			pPlayer = CBaseEntity::Instance( pPlayerEdict );
		}
	}
	
	return pPlayer;
}

